import logging
from multiprocessing import Process, Manager
from rockettm import tasks
import traceback
from kombu import Connection, Exchange, Queue
import sys
import os
from timekiller import call
import importlib
import requests
import time
from basicevents import run, send, subscribe


if len(sys.argv) == 2:
    i, f = os.path.split(sys.argv[1])
    sys.path.append(i)
    settings = __import__(os.path.splitext(f)[0])
else:
    sys.path.append(os.getcwd())
    try:
        import settings
    except:
        exit("settings.py not found")

logging.basicConfig(**settings.logger)


try:
    callback_api = settings.callback_api
except:
    callback_api = None

for mod in settings.imports:
    importlib.import_module(mod)

tasks.ip = settings.ip


@subscribe('api')
def call_api(json):
    if callback_api:
        try:
            requests.post(callback_api, json=json)
        except:
            pass


def safe_worker(func, return_dict, apply_max_time, body):
    try:
        return_dict['result'] = call(func, apply_max_time,
                                     *body['args'], **body['kwargs'])
        return_dict['success'] = True
    except:
        return_dict['result'] = traceback.format_exc()
        return_dict['success'] = False
        logging.error(return_dict['result'])


def worker(name, concurrency, durable=False, max_time=-1):
    def safe_call(func, apply_max_time, body):
        return_dict = Manager().dict()
        p = Process(target=safe_worker, args=(func, return_dict,
                                              apply_max_time, body))
        p.start()
        p.join()
        return return_dict

    def callback(body, message):
        message.ack()
        logging.info("execute %s" % body['event'])
        _id = body['args'][0]
        send('api', {'_id': _id, 'status': 'processing'})
        if not body['event'] in tasks.subs:
            send('api', {'_id': _id,
                         'result': 'task not defined',
                         'status': 'finished',
                         'success': False})
            return False

        result = []
        for func, max_time2 in tasks.subs[body['event']]:
            logging.info("exec func: %s, timeout: %s" % (func, max_time2))
            if max_time2 != -1:
                apply_max_time = max_time2
            else:
                apply_max_time = max_time
            result.append(dict(safe_call(func, apply_max_time,
                                         body)))

        success = not any(r['success'] is False for r in result)
        send('api', {'_id': _id, 'status': 'finished',
                     'success': success, 'result': result})
        return True

    while True:
        try:
            with Connection('amqp://guest:guest@%s//' % settings.ip) as conn:
                conn.ensure_connection()
                exchange = Exchange(name, 'direct', durable=durable)
                queue = Queue(name=name,
                              exchange=exchange,
                              durable=durable, routing_key=name)
                queue(conn).declare()
                logging.info("create queue: %s durable: %s" % (name, durable))
                channel = conn.channel()
                channel.basic_qos(prefetch_size=0, prefetch_count=1,
                                  a_global=False)
                with conn.Consumer(queue, callbacks=[callback],
                                   channel=channel) as consumer:
                    while True:
                        logging.info(consumer)
                        conn.drain_events()

        except (KeyboardInterrupt, SystemExit):
            logging.warning("server stop!")
            break

        except:
            # logging.error(traceback.format_exc())
            logging.error("connection loss, try reconnect")
            time.sleep(5)


def main():
    # start basicevents
    run()
    list_process = []
    for queue in settings.queues:
        for x in range(queue['concurrency']):
            p = Process(target=worker, kwargs=queue)
            logging.info("start process worker: %s queue: %s" % (worker,
                                                                 queue))
            list_process.append(p)
            p.start()

    for p in list_process:
        p.join()

if __name__ == "__main__":
    main()



from kombu import Connection, Exchange, Queue
import logging
import uuid
import traceback
import time


class tasks(object):
    subs = {}
    queues = []
    ip = "localhost"
    conn = False
    producer = False

    # deprecated
    @staticmethod
    def connect(ip=None):
        if ip:
            tasks.ip = ip
        logging.warning('reconnect amqp')
        tasks.conn = Connection('amqp://guest:guest@%s//' % tasks.ip)
        tasks.producer = tasks.conn.Producer(serializer='json')

    @staticmethod
    def add_task(event, func, max_time=-1):
        logging.info("add task %s" % event)
        if event not in tasks.subs:
            tasks.subs[event] = []
        tasks.subs[event].append((func, max_time))

    @staticmethod
    def task(event, max_time=-1):
        def wrap_function(func):
            tasks.add_task(event, func, max_time)
            return func
        return wrap_function

    @staticmethod
    def send_task(queue_name, event, *args, **kwargs):
        if 'rocket_id' in kwargs:
            _id = kwargs.pop('rocket_id')
        else:
            _id = str(uuid.uuid4())

        args = list((_id,) + args)
        logging.info("send task to queue %s, event %s" % (queue_name, event))
        exchange = Exchange(queue_name, 'direct', durable=True)
        queue = Queue(queue_name, exchange=exchange, routing_key=queue_name,
                      durable=True)
        for retry in range(10):
            if not tasks.conn or not tasks.conn.connected:
                tasks.connect()
            try:
                tasks.producer.publish({'event': event, 'args': args,
                                        'kwargs': kwargs},
                                       exchange=exchange,
                                       routing_key=queue_name,
                                       declare=[queue])
                break
            except:
                logging.error(traceback.format_exc())
                time.sleep(retry * 1.34)

        logging.warning("send its ok!")
        return _id


# avoids having to import tasks
connect = tasks.connect
send_task = tasks.send_task
add_task = tasks.add_task
task = tasks.task
